import abc
import logging
from flask import request
from flask_restful import Resource, abort

class BaseResource(Resource):
    @abc.abstractproperty
    def key(self):
        raise NotImplementedError
    
    @abc.abstractproperty
    def resource(self):
        raise NotImplementedError

    def _validate(self, req):
        raise NotImplementedError

class ListResource(BaseResource):
    def _validate(self, req):
        pass
        # self.resource.validate(req)

    def get(self):
        obj_list = self.resource.all()
        obj_dict_list = [obj.to_dict() for obj in obj_list]

        return {self.key: obj_dict_list}

    def post(self):
        req = request.get_json(force=True)
        self._validate(req)

        obj = self.resource.from_dict(req)

        try:
            obj.create()
        except Exception as e:
            logging.error(e)
            abort(500)

        return {self.key: obj.to_dict()}, 200

class CrudResource(BaseResource):
    def _retrieve_by_id(self, object_id):
        try:
            obj = self.resource.query.get(object_id)
        except:
            abort(404)
        return obj

    def _validate(self, req):
        pass
        # return self.resource.validate(req)

    def get(self, obj_id):
        obj = self._retrieve_by_id(obj_id)

        return {self.key: obj.to_dict()}

    def put(self, obj_id):
        obj = self._retrieve_by_id(obj_id)

        req = request.get_json(force=True)
        self._validate(req)
        
        try:
            obj.update(req)
        except Exception as e:
            logging.error(e)
            abort(500)

        return {self.key: obj.to_dict()}
    
    def delete(self, obj_id):
        obj = self._retrieve_by_id(obj_id)

        try:
            obj.delete()
        except Exception as e:
            logging.error(e)
            abort(500)

        return {self.key: obj.to_dict()}



from argparse import ArgumentParser
import os
import sys

from .cmd.init import InitCmd
from .cmd.profile import ProfileCmd
from .cmd.run import RunCmd

# Add other modules to PYTHON_PATH
path = os.path.abspath('../')
sys.path.append(path)


class CliParser(object):
    """Argument parser for the CLI"""

    def __init__(self):
        self._parser = ArgumentParser()
        self._add_arguments()

    def _add_arguments(self):
        """Adds arguments to parser."""
        self._parser.add_argument(
            'command', type=str
        )
        self._parser.add_argument(
            '-v', '--version',
            action='version', 
            version=self.version(),
            help='Show the version number.'
        )

    def parse(self, argv):
        args = self._parser.parse_args(argv[1:])
        getattr(self, args.command)(argv[1:])
        
    # Commands
    def version(self, args=None):
        return format_program_version(get_installation_info().version, sys.version.split()[0])

    def init(self, args):
        args = None
        task = InitCmd.from_args(args)
        task.run()

    def run(self, args):
        args = None
        task = RunCmd.from_args(args)
        task.run()

    def profile(self, args):
        args = None
        task = ProfileCmd.from_args(args)
        task.run()

def get_installation_info():
    import pkg_resources

    return pkg_resources.require('monosi')[0]

def format_program_version(pkg_version, python_version):
    return u'monosi {} using Python {}\n'.format(pkg_version, python_version)




import abc
from dataclasses import dataclass
from typing import List, Type

from sqlalchemy import create_engine

from .column import Column, ColumnDataType


class BaseDialect:
    @classmethod
    def approx_distinct_count(cls):
        return "COUNT(DISTINCT {})"
    
    @classmethod
    def approx_distinctness(cls):
        return "{} / CAST(COUNT(*) AS NUMERIC)".format(cls.approx_distinct_count())

    @classmethod
    def numeric_mean(cls):
        return "AVG({})"

    @classmethod
    def numeric_min(cls):
        return "MIN({})"

    @classmethod
    def numeric_max(cls):
        return "MAX({})"

    @classmethod
    def numeric_std(cls):
        return "STDDEV(CAST({} as double))"

    @classmethod
    def mean_length(cls):
        return cls.numeric_mean().format("LENGTH({})")

    @classmethod
    def max_length(cls):
        return cls.numeric_max().format("LENGTH({})")

    @classmethod
    def min_length(cls):
        return cls.numeric_min().format("LENGTH({})")

    @classmethod
    def std_length(cls):
        return cls.numeric_std().format("LENGTH({})")

@dataclass
class BaseDriverConfiguration:
    @abc.abstractmethod
    def driver_name(self):
        raise NotImplementedError

    @abc.abstractclassmethod
    def configuration_schema(cls):
        raise NotImplementedError

    @abc.abstractmethod
    def connection_string(self) -> str:
        raise NotImplementedError

    @abc.abstractclassmethod
    def validate(cls, config_dict):
        raise NotImplementedError

    @classmethod
    def from_dict(cls, config_dict):
        raise NotImplementedError

    @abc.abstractmethod
    def to_dict(self):
        raise NotImplementedError

@dataclass
class BaseDriver:
    configuration: BaseDriverConfiguration
    dialect: Type[BaseDialect]

    def _retrieve_type(self, type_code, scale):
        raise NotImplementedError

    def _create_columns(self, columns_and_types) -> List[Column]:
        column_names = []
        column_count = {}

        columns = []

        for column_and_type in columns_and_types:
            name = column_and_type[0]
            column_type = column_and_type[1]

            try:
                curr_count = column_count[name] + 1
            except:
                curr_count = 0
            column_count[name] = curr_count

            column_name = "{}{}".format(name, (column_count[name] if curr_count else ''))
            column_names.append(column_name)

            columns.append(Column(column_name.upper(), column_type))

        return columns

    def _retrieve_results(self, cs):
        column_and_types = [(d.name, self._retrieve_type(d.type_code, d.scale)) for d in cs.cursor.description]
        columns = self._create_columns(column_and_types)
        rows = [dict(zip([col.name for col in columns], row)) for row in cs.fetchall()]

        return {
            "columns": columns, 
            "rows": rows,
        }

    def _before_execute(self):
        pass

    @abc.abstractmethod
    def _execute(self, sql):
        raise NotImplementedError

    def execute(self, sql):
        self._before_execute()

        cs = self._execute(sql)
        results = self._retrieve_results(cs)

        return results

class BaseSqlAlchemyDriver(BaseDriver):
    def __init__(self, configuration: BaseDriverConfiguration):
        self.configuration = configuration
        self.connection = self._open()

    def _open(self):
        try:
            engine = create_engine(self.configuration.connection_string())
            return engine.connect()
        except Exception as e:
            raise e

    def _execute(self, sql):
        if not self.connection:
            raise Exception("Connection has already been closed. Could not execute.")

        return self.connection.execute(sql)

    def close(self):
        if self.connection:
            self.connection.close()
        self.connection = None

    @classmethod
    def validate(cls, config_dict):
        pass

@dataclass
class DriverDefinition:
    type: str
    configuration: str # DriverConfig
    name: str = "default"

    def to_dict(self):
        return {
            'name': self.name,
            'type': self.type,
            'configuration': self.configuration,
        }

    @classmethod
    def from_dict(cls, driver_dict):
        return cls(
            name=driver_dict['name'],
            type=driver_dict['type'],
            configuration=driver_dict['configuration'],
        )



from core.common.drivers.column import Column, Table
from core.common.drivers.factory import load_driver

from core.monitor.models.table import TableMonitorDefinition
from core.monitor.models.schema import SchemaMonitorDefinition

class ProfileTask:
    def __init__(self, driver_config):
        self.driver_config = driver_config

    def _retrieve_tables(self):
        driver_cls = load_driver(self.driver_config)
        driver = driver_cls(self.driver_config)
        metadata = driver.metadata()

        return Table.from_metadata(metadata)

    def _create_definitions(self):
        definitions = []

        tables = self._retrieve_tables()
        for table in tables:
            try:
                table_monitor_def = TableMonitorDefinition(
                    name="{} - Table Health".format(table.name),
                    description="Generated by profiler",
                    type="table",
                    configuration='{"table": "' + table.name + '", "timestamp_field": "' + table.timestamp().name + '"}',
                    table=table.name,
                    timestamp_field=table.timestamp().name,
                )
                definitions.append(table_monitor_def)
                schema_monitor_def = SchemaMonitorDefinition(
                    name="{} - Schema".format(table.name),
                    description="Generated by profiler",
                    type="schema",
                    table=table.name,
                    columns=[col.to_dict() for col in table.columns],
                    configuration='{"table": "' + table.name + '"}',
                )
                definitions.append(schema_monitor_def)
            except Exception as e:
                pass
                
        return definitions

    def run(self):
        return self._create_definitions()


from dataclasses import dataclass
import json


@dataclass
class IntegrationDefinition:
    name: str
    configuration: str
    enabled: bool = True

    @classmethod
    def validate(cls, monitor_dict):
        pass

    def to_dict(self):
        return {
            "name": self.name,
            "configuration": json.loads(self.configuration),
            "enabled": self.enabled,
            "type": self.__class__.__name__,
        }

    @classmethod
    def from_dict(cls, integration_dict):
        enabled = integration_dict.get('enabled') or True
        return cls(
            name=integration_dict['name'],
            enabled=enabled,
            configuration=json.dumps(integration_dict['configuration']),
        )
        
    def write(self, message):
        raise NotImplementedError()

    @classmethod
    def configuration_schema(cls):
        return {}



from dataclasses import dataclass
from enum import Enum

class ScheduleType(Enum):
    INTERVAL = 'interval'

@dataclass
class Schedule:
    minutes: int = 720
    type: ScheduleType = ScheduleType.INTERVAL

    def to_dict(self):
        return {
            "minutes": self.minutes,
            "type": self.type._value_,
        }



import requests
import sys
import json

from server.integrations.base import IntegrationDefinition

class SlackIntegration(IntegrationDefinition):
    @classmethod
    def configuration_schema(cls):
        return {
            "type": "object",
            "properties": {
                "url": {"type": "string", "title": "Slack Webhook URL"},
            },
            "secret": [ "url" ],
        }

    @staticmethod
    def _create_request_data(message):
        body = {
            "text": message
        }
        byte_length = str(sys.getsizeof(body))
        headers = {'Content-Type': "application/json", 'Content-Length': byte_length}
        return headers, json.dumps(body)

    def write(self, message, color=None):
        url = json.loads(self.configuration)['url']
        headers, data = self._create_request_data(message)

        try:
            response = requests.post(url, data=data, headers=headers)

            if response.status_code != 200:
                raise Exception(response.status_code, response.text)
        except Exception as e:
            print(e)
            raise Exception("Request to Slack webhook URL {} failed.".format(url))

    def dump_failures(self, failed_tests):
        pass
        # if len(failed_tests) > 0:
        #     message = ""
        #     message += "\nFailures\n"

        #     for failed_test in failed_tests:
        #         message += "\tAnomalies: {}\n".format(len(failed_test.anomalies()))
        #         message += "\n"

        #     self.write(message)

    def dump_summary(self, summary):
        monitor = summary['monitor']

        message = ""
        message += "Monitor: {monitor_name}".format(monitor_name=monitor.name)
        if monitor.description:
            message += "\nDescription: {monitor_description}".format(monitor_description=monitor.description)
        message += "\nTotal Time: {} seconds, SQL Load Time: {} seconds".format(summary['total_time'], summary['load_time'])
        message += "\n"
        
        message += "\nTotal Metrics: {}".format(summary['test_count'])
        message += "\nFailed Metrics: {}".format(summary['failed_count'])
        message += "\n\nAnomaly details coming soon to slack integration."
        self.write(message)

    def dump_pending(self, none_obj):
        pass

    def close(self, none_obj):
        pass

    def monitor_started(self, none_obj):
        pass

    def monitor_finished(self, none_obj):
        pass

    def message(self, none_obj):
        pass

    def stop(self, none_obj):
        pass

    def dump_skipped(self, skipped_tests):
        pass

    def test_started(self, none_obj):
        pass

    def test_finished(self, none_obj):
        pass

    def test_passed(self, none_obj):
        pass
        
    def test_pending(self, none_obj):
        pass

    def test_failed(self, none_obj):
        pass

    def start_dump(self, none_obj):
        # self.write('start_dump', Color.ENDC)
        pass




from server.models.integration import Integration
from .base import CrudResource, ListResource

class IntegrationListResource(ListResource):
    @property
    def resource(self):
        return Integration

    @property
    def key(self):
        return "integrations"

class IntegrationResource(CrudResource):
    @property
    def resource(self):
        return Integration

    @property
    def key(self):
        return "integration"




from flask_sqlalchemy import BaseQuery, SQLAlchemy

from .models import Base

class MsiQuery(BaseQuery):
    def get_by_id(cls, ident, default=None):
        return self.get(ident) or default

    # def all(cls):
    #     return cls.query.all()
    
    # def create(self):
    #     db.session.add(self)
    #     db.session.commit()

    # def update(self, updates):
    #     for k, v in updates.items():
    #         setattr(self, k, v)
    #     db.session.commit()

    # def delete(self):
    #     db.session.delete(self)
    #     db.session.commit()

db = SQLAlchemy(model_class=Base, query_class=MsiQuery)




from dataclasses import dataclass, field
from enum import Enum

from typing import Any, List, Dict
import json
import operator as py_operator

from core.common.drivers.base import BaseDialect

from . import MonitorConfiguration, MonitorDefinition
from .base import Monitor
from .metrics import MetricBase, MetricType
from .schedule import Schedule

# class Operator(Enum):
#     EQ = 'eq'
#     NE = 'ne'
#     GT = 'gt'
#     GE = 'ge'
#     LT = 'lt'
#     LE = 'le'
#     ABS_INC = 'abs_inc'
#     ABS_dec = 'abs_dec'
#     REL_INC = 'rel_inc'
#     REL_dec = 'rel_dec'

# class ThresholdDefinition:
# 	operator: Operator
# 	value: float

@dataclass
class CustomMonitorConfiguration:
    sql: str
    thresholds: str
# 	thresholds: List[ThresholdDefinition]

@dataclass
class CustomMonitorDefinition(MonitorConfiguration, MonitorDefinition, CustomMonitorConfiguration):
    def to_monitor(self, workspace):
        return CustomMonitor.from_definition(self, workspace)


class ComparativeOperatorDefaults:
    EQ = 'eq'
    NE = 'ne'
    GT = 'gt'
    GE = 'ge'
    LT = 'lt'
    LE = 'le'

class PercentOperatorDefaults:
    ABS_INC = 'abs_inc'
    ABS_dec = 'abs_dec'
    REL_INC = 'rel_inc'
    REL_dec = 'rel_dec'

class Operator(ComparativeOperatorDefaults, PercentOperatorDefaults, Enum):
    EQ = 'eq'
    NE = 'ne'
    GT = 'gt'
    GE = 'ge'
    LT = 'lt'
    LE = 'le'
    ABS_INC = 'abs_inc'
    ABS_dec = 'abs_dec'
    REL_INC = 'rel_inc'
    REL_dec = 'rel_dec'

    def fn(self):
        if self._value_ in ComparativeOperatorDefaults.__dict__.values():
            return getattr(py_operator, self._value_)
        elif self._value_ in PercentOperatorDefaults.__dict__.values():
            return py_operator.iadd # TODO: Update
        
        raise Exception("Operator does not exist.")

@dataclass
class Threshold:
    operator: Operator
    value: float

    def evaluate(self, source_operand: float):
        op_fn = self.operator.fn()
        return op_fn(source_operand, self.value)

    @classmethod
    def from_dict(cls, value: Dict[str, Any]) -> 'Threshold':
        return cls(
            operator=Operator(value['operator']),
            value=value['value'],
        )

    def to_dict(self):
    	return {
    		'operator': self.operator._value_,
    		'value': self.value,
    	}

@dataclass
class CustomMetric(MetricBase):
    sql: str = field(default_factory=str)
    thresholds: List[Threshold] = field(default_factory=list)
    type: MetricType = MetricType.CUSTOM

    @classmethod
    def from_dict(cls, value: Dict[str, Any]) -> 'CustomMetric':
        sql = value['sql']
        thresholds: List[Threshold] = []

        for threshold_dict in value['thresholds']:
            threshold = Threshold.from_dict(threshold_dict)
            thresholds.append(threshold)

        return cls(
            sql=sql, 
            thresholds=thresholds,
        )

    def to_dict(self):
    	return {
    		'sql': self.sql,
    		'thresholds': [threshold.to_dict() for threshold in self.metric.thresholds],
    	}

    def compile(self, dialect: BaseDialect):
        return self.sql

def extract_or_default(obj, key, default):
    return obj[key] if key in obj else default

@dataclass
class CustomMonitor(Monitor):
    metric: CustomMetric

    def retrieve_metrics(self):
        return [self.metric]

    def info(self):
        info_str = "Custom Monitor"
        if self.description:
            info_str += ": {}".format(self.description)
        return info_str

    @classmethod
    def validate(cls, monitor_dict):
        pass # only one CustomMetric allowed per monitor
    
    @classmethod
    def from_dict(cls, value: Dict[str, Any]) -> 'CustomMonitor':
        metric = CustomMetric.from_dict(value)
        description = extract_or_default(value, 'description', None)
        schedule = Schedule.from_dict(extract_or_default(value, 'schedule', {}))

        return cls(
            metrics=[metric],
            description=description,
            schedule=schedule,
        )

    def to_dict(self):
        metric_dict = self.metric.to_dict()
        return {
            'sql': metric_dict['sql'],
            'thresholds': metric_dict['thresholds'],
            'type': 'custom',
        }

    def base_sql_statement(self, select_sql, dialect):
        return "WITH custom_metric AS ({select_sql}) SELECT 1 AS custom FROM custom_metric".format(select_sql=select_sql)

    @classmethod
    def from_definition(cls, definition: CustomMonitorDefinition, workspace):
        # monitor_base = super().from_definition(definition, workspace)
        driver_config = workspace.get_driver_config(definition.datasource)
        metric = CustomMetric.from_dict({'sql': definition.sql, 'thresholds': definition.thresholds})

        return cls(
            # name=monitor_base.name,
            # description=monitor_base.description,
            # enabled=monitor_base.enabled,
            # driver_config=monitor_base.driver_config,
            name=definition.name,
            description=definition.description,
            schedule=Schedule(definition.schedule_minutes),
            enabled=definition.enabled or True,
            driver_config=driver_config,
            metric=metric,
        )



from dataclasses import dataclass
from sqlalchemy import Boolean, Column, DateTime, Integer, String, Table, Text, func

from server.integrations import reporter
from server.integrations.base import IntegrationDefinition
from server.integrations.slack import SlackIntegration

from . import Base, mapper_registry
from .base import CrudMixin


# TODO: Is this necessary? Is it actually mapped?
# @mapper_registry.mapped

@dataclass
class Integration(IntegrationDefinition, Base, CrudMixin):
    # TODO: Abstract id, created_at, updated_at
    __table__ = Table(
        "integration",
        mapper_registry.metadata,
        Column("id", Integer, primary_key=True),

        Column("name", String(50)),
        Column("type", String(50)), # TODO: Enum
        Column("enabled", Boolean),
        Column("configuration", Text), # TODO: Better way to not store JSON-style configuration?

        # TODO: Add updated_on fn
        Column("updated_at", DateTime(timezone=True), nullable=False, server_default=func.now()),
        Column("created_at", DateTime(timezone=True), nullable=False, server_default=func.now()),
    )

    def to_dict(self):
        obj_dict = super().to_dict()

        obj_dict['id'] = self.id
        obj_dict['updated_at'] = str(self.updated_at)
        obj_dict['created_at'] = str(self.created_at)

        return obj_dict

    def register(self):
        # TODO: Support more than slack
        integration = SlackIntegration(
            name=self.name,
            enabled=self.enabled,
            configuration=self.configuration,
        )
        reporter.register_listener(integration)

    def create(self):
        super().create()
        self.register()



import os

from cli.config import ProjectConfiguration
from cli.utils.yaml import write_file

class InitCmd(object):
    @classmethod
    def from_args(cls, args=None):
        return cls()

    def run(self):
        project_name = "monosi-repo"
        project_directory = os.path.join(os.getcwd(), project_name)

        if os.path.exists(project_name):
            print("Directory {} already found. Could not initialize.".format(project_name))
            return

        args = None
        project_configuration = ProjectConfiguration(
            root_path=os.path.abspath(os.getcwd()),
            project_name="monosi-repo",
            workspace_name="default",
        )
        os.mkdir(project_directory)

        project_file = os.path.join(project_directory, 'monosi_project.yml')
        write_file(project_file, project_configuration.to_dict())

        monitors_dir = os.path.join(project_directory, 'monitors')
        if not os.path.exists(monitors_dir):
            os.mkdir(monitors_dir)
            monitor_example_path = os.path.join(monitors_dir, 'example.yml')
            file_contents = {
                'monosi': {
                    'monitors': [
                        {
                            'name': 'Example Monitor',
                            'description': 'Monitor created using the `monosi init` command',
                            'type': 'custom',
                            'sql': 'SELECT 1 + 1',
                            'thresholds': [
                                {
                                    'operator': 'eq',
                                    'value': 2,
                                },
                            ],
                        }
                    ]
                }
            }
            write_file(monitor_example_path, file_contents)

        print("Successfully initialized monosi project: {}".format(project_name))



from typing import Type

from core.common.drivers.base import BaseDialect, BaseSqlAlchemyDriver
from core.common.drivers.column import ColumnDataType

from .configuration import PostgresDriverConfiguration
from .dialect import PostgresDialect


POSTGRES_SYSTEM_TABLES = [
    'pg_aggregate',
    'pg_am',
    'pg_amop',
    'pg_amproc',
    'pg_attrdef',
    'pg_attribute',
    'pg_authid',
    'pg_auth_members',
    'pg_cast',
    'pg_class',
    'pg_constraint',
    'pg_collation',
    'pg_conversion',
    'pg_database',
    'pg_db_role_setting',
    'pg_default_acl',
    'pg_depend',
    'pg_description',
    'pg_enum',
    'pg_extension',
    'pg_foreign_data_wrapper',
    'pg_foreign_server',
    'pg_foreign_table',
    'pg_index',
    'pg_inherits',
    'pg_language',
    'pg_largeobject',
    'pg_largeobject_metadata',
    'pg_namespace',
    'pg_opclass',
    'pg_operator',
    'pg_opfamily',
    'pg_pltemplate',
    'pg_proc',
    'pg_rewrite',
    'pg_seclabel',
    'pg_shdepend',
    'pg_shdescription',
    'pg_statistic',
    'pg_tablespace',
    'pg_trigger',
    'pg_ts_config',
    'pg_ts_config_map',
    'pg_ts_dict',
    'pg_ts_parser',
    'pg_ts_template',
    'pg_type',
    'pg_user_mapping',
    # System Views
    'pg_available_extensions',
    'pg_available_extension_versions',
    'pg_cursors',
    'pg_group',
    'pg_indexes',
    'pg_locks',
    'pg_prepared_statements',
    'pg_prepared_xacts',
    'pg_roles',
    'pg_rules',
    'pg_seclabels',
    'pg_settings',
    'pg_shadow',
    'pg_stats',
    'pg_tables',
    'pg_timezone_abbrevs',
    'pg_timezone_names',
    'pg_user',
    'pg_user_mappings',
    'pg_views',

    'information_schema',
    'pg_catalog',
]

POSTGRES_TYPES = {
    
}

class PostgresDriver(BaseSqlAlchemyDriver):
    configuration: PostgresDriverConfiguration
    dialect: Type[BaseDialect] = PostgresDialect

    def _retrieve_type(self, type_code, scale):
        resolved_type = POSTGRES_TYPES.get(type_code, None)
        if resolved_type == ColumnDataType.STRING and scale is not None and scale > 0:
            resolved_type = ColumnDataType.FLOAT

        return resolved_type

    @staticmethod
    def _filter_metadata(results):
        filtered_results = {}
        filtered_results['columns'] = results['columns']
        filtered_results['rows'] = []
        rows = results['rows']
        for row in rows:
            if row['NAME'].lower() not in POSTGRES_SYSTEM_TABLES:
                filtered_results['rows'].append(row)

        return filtered_results

    def metadata(self):
        DATABASE_METADATA_SQL = """
        SELECT
            lower(c.table_name) AS NAME,
            lower(c.column_name) AS COL_NAME,
            lower(c.data_type) AS COL_TYPE,
            ordinal_position AS COL_SORT_ORDER, 
            lower(c.table_catalog) AS DATABASE,
            lower(c.table_schema) AS SCHEMA
        FROM
            INFORMATION_SCHEMA.COLUMNS AS c
        LEFT JOIN
            INFORMATION_SCHEMA.TABLES t
                ON c.TABLE_NAME = t.TABLE_NAME
                AND c.TABLE_SCHEMA = t.TABLE_SCHEMA;
        """

        results = self.execute(DATABASE_METADATA_SQL)
        filtered_results = self._filter_metadata(results)

        return filtered_results




from cli.config import ProjectConfiguration
from core.monitor.tasks.profile import ProfileTask
import os

from cli.utils import yaml

from .base import BaseCmd


BOOTSTRAPPED_MONITOR_PATH = './bootstrapped-monitors'

def _write_definition(definition, monitors_dir=BOOTSTRAPPED_MONITOR_PATH):
    table_parts = definition.table.split('.')
    database = table_parts[0]
    schema = table_parts[1]
    table = table_parts[2]

    monitor_path = os.path.join(monitors_dir, database, schema)
    if not os.path.exists(monitor_path):
        os.makedirs(monitor_path)

    path = os.path.join(monitor_path, table + '.yml')
    if os.path.exists(path):
        try:
            file_contents = yaml.parse_file(path)
            file_contents['monosi']['monitors'].append(definition.to_dict())
        except Exception as e:
            file_contents = {'monosi': {'monitors': definition.to_dict()}}
    else:
        file_contents = {'monosi': {'monitors': [definition.to_dict()]}}
    yaml.write_file(path, file_contents)

def _persist_definitions(definitions):
    if not os.path.exists(BOOTSTRAPPED_MONITOR_PATH):
        os.makedirs(BOOTSTRAPPED_MONITOR_PATH)

    for definition in definitions:
        _write_definition(definition)

class ProfileCmd(BaseCmd):
    def _create_tasks(self):
        sources = self.project.configuration.sources.values()

        return [ProfileTask(source) for source in sources]

    def _process_tasks(self):
        results = [task.run() for task in self.task_queue]
        for result in results:
            _persist_definitions(result)

        self.project.configuration.add_monitor_path('./bootstrapped-monitors')
        yaml.write_file(os.path.join(os.getcwd(), 'monosi_project.yml'), self.project.configuration.to_dict()) # TODO: might be yaml

        print("Profiling complete.")



import os

def monosi_home_dir():
    monosi_path = os.path.expanduser("~/.monosi")
    if not os.path.exists(monosi_path):
        os.makedirs(monosi_path)

    return monosi_path

class BaseConfig:
    """Base configuration."""
    SECRET_KEY = os.getenv('SECRET_KEY', 'my_precious')
    DEBUG = False
    SQLALCHEMY_DATABASE_URI = "sqlite:///" + monosi_home_dir() + "/sqlite.db"
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    SERVE_UI = bool(os.getenv('SERVE_UI', False))

class DevelopmentConfig(BaseConfig):
    """Development configuration."""
    DEBUG = True

class TestingConfig(BaseConfig):
    """Testing configuration."""
    DEBUG = True
    TESTING = True

class ProductionConfig(BaseConfig):
    """Production configuration."""
    SECRET_KEY = 'my_precious'
    DEBUG = False

Config = DevelopmentConfig
if os.getenv('FLASK_ENV') == "production":
    Config = ProductionConfig



from dataclasses import dataclass
from typing import Type

from core.common.drivers.base import BaseDialect, BaseSqlAlchemyDriver
from core.common.drivers.column import ColumnDataType

from .configuration import SnowflakeDriverConfiguration
from .dialect import SnowflakeDialect


SNOWFLAKE_TYPES = {
    0: ColumnDataType.INTEGER,
    1: ColumnDataType.FLOAT,
    2: ColumnDataType.STRING,
    3: ColumnDataType.DATE,
    4: ColumnDataType.DATETIME,
    5: ColumnDataType.STRING,
    6: ColumnDataType.DATETIME,
    7: ColumnDataType.DATETIME,
    8: ColumnDataType.DATETIME,
    13: ColumnDataType.BOOLEAN,
}

SNOWFLAKE_SYSTEM_TABLES = [
    'databases',
    'external_tables',
    'file_formats',
    'functions',
    'load_history',
    'object_privileges',
    'table_privileges',
    'pipes',
    'procedures',
    'referential_constraints',
    'replication_databases',
    'schemata',
    'sequences',
    'stages',
    'table_constraints',
    'table_storage_metrics',
    'tables',
    'usage_privileges',
    'views'
]


@dataclass
class SnowflakeDriver(BaseSqlAlchemyDriver):
    def __init__(self, configuration: SnowflakeDriverConfiguration):
        self.configuration: SnowflakeDriverConfiguration = configuration
        self.dialect: Type[BaseDialect] = SnowflakeDialect
        self.connection = self._open()

    def _retrieve_type(self, type_code, scale):
        resolved_type = SNOWFLAKE_TYPES.get(type_code, None)
        if resolved_type == ColumnDataType.STRING and scale is not None and scale > 0:
            resolved_type = ColumnDataType.FLOAT

        return resolved_type

    def _before_execute(self):
        if not self.connection:
            return

        warehouse_sql = 'USE WAREHOUSE {}'.format(self.configuration.warehouse)
        database_sql = 'USE {}'.format(self.configuration.database)

        try:
            self.connection.execute(warehouse_sql)
            self.connection.execute(database_sql)
        except Exception as e:
            raise e

    @staticmethod
    def _filter_metadata(results):
        filtered_results = {}
        filtered_results['columns'] = results['columns']
        filtered_results['rows'] = []
        rows = results['rows']
        for row in rows:
            if row['NAME'].lower() not in SNOWFLAKE_SYSTEM_TABLES:
                filtered_results['rows'].append(row)

        return filtered_results

    def metadata(self):
        DATABASE_METADATA_SQL = """
        SELECT
            lower(c.table_name) AS name,
            lower(c.column_name) AS col_name,
            lower(c.data_type) AS col_type,
            c.comment AS col_description,
            lower(c.ordinal_position) AS col_sort_order,
            lower(c.table_catalog) AS database,
            lower(c.table_schema) AS schema,
            t.comment AS description,
            decode(lower(t.table_type), 'view', 'true', 'false') AS is_view
        FROM
            {database}.INFORMATION_SCHEMA.COLUMNS AS c
        LEFT JOIN
            {database}.INFORMATION_SCHEMA.TABLES t
                ON c.TABLE_NAME = t.TABLE_NAME
                AND c.TABLE_SCHEMA = t.TABLE_SCHEMA;
        """.format(database=self.configuration.database)

        results = self.execute(DATABASE_METADATA_SQL)
        filtered_results = self._filter_metadata(results)

        return filtered_results




from typing import Iterable

from core.monitor.models import MonitorDefinition
from core.monitor.models.custom import CustomMonitorDefinition
from core.monitor.models.schema import SchemaMonitorDefinition
from core.monitor.models.table import TableMonitorDefinition
from cli.utils.files import File

from . import YamlParser

# TODO: Replace with configuration object
def monitor_definition_factory(monitor_dict):
    monitor_type = monitor_dict['type'].lower()
    if monitor_type == 'table':
        return TableMonitorDefinition
    elif monitor_type == 'custom':
        return CustomMonitorDefinition
    elif monitor_type == 'schema':
        return SchemaMonitorDefinition

    raise Exception("Could not find monitor type: {}".format(monitor_type))

class MonitorParser(YamlParser):
    def __init__(self, configuration):
        super().__init__(configuration, ['monosi', 'monitors'])

    def parse_monitors(self, file: File) -> Iterable[MonitorDefinition]:
        monitors_dict = self.extract_from_file(file)

        for monitor_dict in monitors_dict:
            try:
                monitor_def_class = monitor_definition_factory(monitor_dict)
                monitor_def_class.validate(monitor_dict)
                monitor = monitor_def_class.from_dict(monitor_dict)
            except Exception as e:
                raise e
            yield monitor

    def parse_file(self, file, project):
        for monitor in self.parse_monitors(file):
            project.add_monitor(monitor)



import abc
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

from core.common.drivers.column import Table

from core.monitor.models.metrics import MetricBase

@dataclass
class DataPoint:
    value: Any
    error: bool = False

    def to_dict(self):
        return {
            'value': self.value,
            'error': self.error,
        }

@dataclass
class Data(Table):
    points: Dict[str, List[DataPoint]]

    def for_metric(self, metric: MetricBase) :
        alias = metric.alias().lower()
        return self.points[alias]

    @classmethod
    def from_results(cls, results):
        cols = [column.name.lower() for column in results['columns']]
        points = {}
        for col in cols:
            points[col] = []

        for row in results['rows']:
            for col in row.keys():
                if row[col]:
                    try:
                        point = DataPoint(value=row[col])
                        points[col.lower()].append(point)
                    except TypeError:
                        pass
                        # logging.info("Can't convert other rows, including window_start for example.")

        try:
            tables = Table.from_metadata(results)
            if len(tables) != 1:
                raise Exception("Schema parsing failed, more than 1 or no tables.")
            columns = tables[0].columns
        except:
            columns = results['columns']
            # this is special casing for schema 
        
        return cls(
            name="intermediate-data",
            columns=columns,
            points=points
        )

@dataclass
class Anomaly:
    points: List[DataPoint] = field(default_factory=list)

@dataclass
class TestResult:
    anomalies: List[Anomaly]

@dataclass
class Test:
    data: List[DataPoint]

    def anomalies(self):
        anomaly_pts = list(filter(lambda x: x.error == False, self.data))
        anomaly_objs = [Anomaly(points=[pt]) for pt in anomaly_pts]
        return anomaly_objs

    @abc.abstractmethod
    def run(self):
        raise NotImplementedError

    @classmethod
    def from_metric(cls, metric: MetricBase, data: Data):
        raise NotImplementedError




import abc
import logging
from flask import request
from flask_restful import Resource, abort

from scheduler.models.execution import Execution


class ExecutionsListResource(Resource):
    def get(self):
        executions = Execution.all()
        executions_dict_list = [execution.to_dict() for execution in executions]

        return {"executions": executions_dict_list}

    # TODO: Execute on demand

class ExecutionsResource(Resource):
    def _retrieve_by_id(self, execution_id):
        try:
            execution = Execution.query.get(execution_id)
        except:
            abort(404)
        return execution

    def get(self, execution_id):
        execution = self._retrieve_by_id(execution_id)

        return {"execution": execution.to_dict()}



import os
import logging
import uuid
from dataclasses import dataclass, field
from typing import Any, Dict, List, Tuple

from core.common.drivers.base import BaseDriverConfiguration
from core.common.drivers.factory import load_config
import core.common.events as events

from cli.utils.files import file_find
import cli.utils.yaml as yaml


DEFAULT_WORKSPACES_DIR = os.path.expanduser('~/.monosi')

def read_user_id(user_filepath: str):
    user_data = yaml.parse_file(user_filepath)
    return user_data['id']

def write_user_id(user_filepath: str):
    user_id = str(uuid.uuid4())

    if not os.path.exists(user_filepath):
        user_data = {"id": user_id}
        yaml.write_file(user_filepath, user_data)

    return user_id

def convert_to_bool(val):
    if val and val.lower() == "true":
        return True
    
    return False

@dataclass
class WorkspaceConfiguration:
    sources: Dict[str, BaseDriverConfiguration] = field(default_factory=list())
    name: str = "default"
    send_anonymous_stats: bool = True
    workspaces_dir: str = DEFAULT_WORKSPACES_DIR # TODO: Update to args
    
    @classmethod
    def _config_from_source(cls, source_dict: Dict[str, Any]):
        from core.common.drivers.factory import load_config

        if 'type' not in source_dict:
            raise Exception("Source type is required.")

        driver_type = source_dict.pop('type')
        try:
            cls = load_config(driver_type)
            config = cls.from_dict(source_dict)
        except Exception as e:
            raise e

        return config

    def validate(self):
        pass

    @classmethod
    def from_dict(cls, workspace_dict: Dict[str, Any]):
        sources = {}
        if 'sources' in workspace_dict:
            for source_name in workspace_dict['sources'].keys():
                source = cls._config_from_source(workspace_dict['sources'][source_name])
                sources[source_name] = source

        config = cls(
            sources=sources,
        )
        config._initialize_events() # TODO: Read value

        return config

    def _initialize_events(self):
        if not self.send_anonymous_stats:
            return

        user_filepath = os.path.join(self.workspaces_dir, '.cookie.yml')
        try:
            if os.path.exists(user_filepath):
                user_id = read_user_id(user_filepath)
            else:
                user_id = write_user_id(user_filepath)
            events.set_user_id(user_id)
        except Exception as e:
            logging.error("There was an issue sending anonymous usage stats with a user id.")
        
    @classmethod
    def _get_workspace_dict(cls, workspace_name: str, all_workspaces_dict: Dict[str, Any]):
        if workspace_name not in all_workspaces_dict:
            raise Exception("Workspace not found: {}".format(workspace_name))

        return all_workspaces_dict[workspace_name]

    @classmethod
    def from_args(cls, args=None) -> 'WorkspaceConfiguration':
        try:
            workspaces_path = file_find(DEFAULT_WORKSPACES_DIR, 'workspaces.y*ml')
        except:
            raise Exception("Could not find workspaces configuration file.")

        workspace_name = "default"
        # TODO: Should check args or project config for a workspace name
        all_workspaces_dict = yaml.parse_file(workspaces_path)
        workspace_dict = cls._get_workspace_dict(workspace_name, all_workspaces_dict)

        return cls.from_dict(workspace_dict)

@dataclass 
class ProjectConfiguration:
    project_name: str
    root_path: str
    workspace_name: str
    # source_name: str
    version: str = '0.0.3' # TODO: Pull from one location
    monitor_paths: List[str] = field(default_factory=lambda: ['./monitors'])
    
    @classmethod
    def from_root_path(cls, root_path: str) -> 'ProjectConfiguration':
        project_config_path = file_find(root_path, 'monosi_project.y*ml')
        project_dict = yaml.parse_file(project_config_path)

        return cls.from_dict(
            root_path=root_path,
            project_dict=project_dict,
        )

    @classmethod
    def validate(cls, project_dict):
        pass

    @classmethod
    def from_dict(
        cls,
        root_path: str,
        project_dict: Dict[str, Any],
    ):
        try:
            cls.validate(project_dict)
        except Exception as e:
            raise e

        project_name = str(project_dict.get('name'))
        version = project_dict.get('version') or '0.0.3'
        workspace_name = project_dict.get('workspace') or 'default'
        source_name = project_dict.get('source')
        log_path: str = project_dict.get('log-path') or 'logs'

        monitor_paths: List[str] = project_dict.get('monitor-paths') or []

        project = cls(
            project_name=project_name,
            version=version,
            root_path=root_path,
            workspace_name=workspace_name,
            # source_name=source_name,
            monitor_paths=monitor_paths,
        )

        return project

    def to_dict(self):
        config_dict = {
            "name": self.project_name,
            "version": self.version,
            "monitor-paths": self.monitor_paths,
        }
        # if self.collection_name:
        #     config_dict.update({"collection": self.collection_name})
        # if self.source_name:
        #     config_dict.update({"source": self.source_name})

        return config_dict


@dataclass
class Configuration(WorkspaceConfiguration, ProjectConfiguration):
    def __post_init__(self):
        try:
            Configuration.validate(self)
        except Exception as e:
            raise e

    @classmethod
    def validate(cls, self):
        pass

    @classmethod
    def create_subclasses(cls, args: Any) -> Tuple[ProjectConfiguration, WorkspaceConfiguration]:
        root_path = os.getcwd()
        if hasattr(args, 'project_dir'):
            root_path = args.project_dir

        project = ProjectConfiguration.from_root_path(root_path)

        # workspace_name = (project.workspace_name or 'default')
        # source_name = (project.source_name or 'default')
        workspace = WorkspaceConfiguration.from_args(
            # workspace_name=workspace_name, 
            # source_name=source_name,
            args=args
        )

        return (project, workspace)

    @classmethod
    def from_subclasses(cls, project: ProjectConfiguration, workspace: WorkspaceConfiguration, args) -> 'Configuration':
        return cls(
            project_name=project.project_name,
            version=project.version,
            root_path=project.root_path,
            monitor_paths=project.monitor_paths,
            workspace_name=project.workspace_name,
            sources=workspace.sources,
            # source_name=project.source_name,
            # reporter=project.reporter,
            # config=workspace.config,
            # send_anonymous_stats=workspace.send_anonymous_stats,
            # args=args
        )

    @classmethod
    def from_args(cls, args: Any) -> 'Configuration':
        project, workspace = cls.create_subclasses(args)
        configuration = cls.from_subclasses(
            project=project,
            workspace=workspace,
            args=args
        )

        return configuration

    def project_dict(self):
        config_dict = {
            "name": self.project_name,
            "version": self.version,
            "monitor-paths": self.monitor_paths,
        }
        if self.workspace_name:
            config_dict.update({"workspace": self.workspace_name})
        # if self.source_name:
        #     config_dict.update({"source": self.source_name})

        return config_dict

    def add_monitor_path(self, path):
        if path not in self.monitor_paths:
            self.monitor_paths.append(path)
            
            project_config_path = file_find(self.root_path, "monosi_project.y*ml")
            project_dict = self.project_dict()

            yaml.write_file(project_config_path, project_dict)

    def get_driver_config(self, source_name: str):
        if source_name not in self.sources.keys():
            raise Exception("Source {} was not defined or can not be found.")

        return self.sources[source_name]



from dataclasses import dataclass
from enum import Enum
from typing import Optional

from core.common.drivers.base import BaseDriverConfiguration

from .schedule import Schedule

@dataclass
class Monitor:
    name: str
    description: Optional[str]
    enabled: bool
    driver_config: BaseDriverConfiguration
    schedule: Schedule
    # type via polymorphism

    def run(self):
        from core.monitor.tasks.run import RunMonitorTask

        return RunMonitorTask(self).run()

    @classmethod
    def from_definition(cls, definition, workspace):
        # MonitorDefinition.validate(definition) # TOOD:  validate the definition
        driver_config = workspace.get_driver_config(definition.datasource)

        return cls(
            name=definition.name,
            description=definition.description,
            enabled=definition.enabled,
            driver_config=driver_config,
        )

    def retrieve_metrics(self):
        raise NotImplementedError




import abc
from dataclasses import dataclass
from enum import Enum

from core.common.drivers.base import BaseDialect

class MetricType(Enum):
    CUSTOM = 'custom'

@dataclass
class MetricBase:
    type: MetricType

    def alias(self):
        return self.type._value_

    @abc.abstractmethod
    def compile(self, dialect: BaseDialect):
        raise NotImplementedError




from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Optional

from core.common.drivers.base import BaseDialect
from core.common.drivers.column import Column

from . import MonitorConfiguration, MonitorDefinition
from .base import Monitor
from .metrics import MetricBase
from .schedule import Schedule


@dataclass
class SchemaMonitorConfigurationDefaults:
	columns: List[Dict[str, str]] = field(default_factory=list)

@dataclass
class SchemaMonitorConfiguration:
    table: str

@dataclass
class SchemaMonitorDefinition(MonitorConfiguration, SchemaMonitorConfigurationDefaults, MonitorDefinition, SchemaMonitorConfiguration):
    def to_monitor(self, workspace):
        return SchemaMonitor.from_definition(self, workspace)

    def to_dict(self):
        return {
            "name": self.name,
            "description": self.description,
            "type": "schema",
            "table": self.table,
            "columns": self.columns,
            # "schedule_minutes": self.schedule_minutes,
            # "schedule_type": self.schedule_type,
            # "schedule": Schedule(self.schedule_minutes).to_dict(),
        }

def extract_or_default(obj, key, default):
    return obj[key] if key in obj else default

class SchemaMetricType(Enum):
    COLUMN_NAMES = 'name'
    COLUMN_TYPES = 'data_type'
    COLUMN_ORDERS = 'order'

    @classmethod
    def all(cls):
        return [
            cls.COLUMN_NAMES,
            cls.COLUMN_TYPES,
            cls.COLUMN_ORDERS,
        ]

@dataclass
class SchemaMetric(MetricBase):
    type: SchemaMetricType
    col_to_metric: Dict[str, Any]

    @classmethod
    def from_columns(cls, metric_type: SchemaMetricType, columns: List[Column]):
        col_to_metric_map = {}
        for col in columns:
            col_to_metric_map[col.name] = getattr(col, metric_type._value_)

        return cls(
            type=metric_type,
            col_to_metric=col_to_metric_map,
        )

    def compile(self, dialect: BaseDialect):
        return ""

@dataclass
class SchemaMonitor(SchemaMonitorConfigurationDefaults, Monitor, SchemaMonitorConfiguration):
    columns: List[Column] = field(default_factory=list)
    metrics: List[SchemaMetric] = field(default_factory=list)

    def retrieve_metrics(self):
        return self.metrics

    def info(self):
        info_str = "Schema Monitor: {}".format(self.table)
        if self.description:
            info_str += ": {}".format(self.description)
        return info_str

    @classmethod
    def validate(cls, monitor_dict):
        if 'table' not in monitor_dict:
            raise Exception("Table key was missing in the schema monitor definition.")
        if 'columns' not in monitor_dict:
            raise Exception("Columns key was missing in the schema monitor definition.")

        if len(monitor_dict['table'].split('.')) < 3:
            raise Exception("Database and schema not found.")

    @classmethod
    def _create_metrics(cls, columns: List[Column]):
        metrics = []
        for metric_type in SchemaMetricType.all():
            metric = SchemaMetric.from_columns(metric_type, columns)
            metrics.append(metric)

        return metrics

    
    @classmethod
    def from_dict(cls, value: Dict[str, Any]) -> 'SchemaMonitor':
        table = value['table'] # Required
        columns = [Column.from_dict(col) for col in value['columns']]
        metrics = cls._create_metrics(columns)

        description = extract_or_default(value, 'description', None)
        schedule = Schedule.from_dict(extract_or_default(value, 'schedule', {}))

        return cls(
            table=table,
            columns=columns,
            metrics=metrics,
            description=description,
            schedule=schedule,
        )

    def to_dict(self):
        return {
            'type': 'schema',
            'table': self.table,
            'columns': [col.to_dict() for col in self.columns],
        }

    def database_name(self):
        table_parts = self.table.split('.')
        if len(table_parts) < 3:
            raise Exception("Table name not fully qualified with database")

        return table_parts[0].lower()

    def schema_name(self):
        table_parts = self.table.split('.')
        if len(table_parts) < 3:
            raise Exception("Schema name not fully qualified with database")

        return table_parts[1].lower()

    def table_name(self):
        table_parts = self.table.split('.')
        if len(table_parts) < 3:
            raise Exception("Table name not fully qualified with database")

        return table_parts[2].lower()

    def base_sql_statement(self, select_sql, dialect):
        return dialect.metadata_query().format(
                database_name=self.database_name(),
                schema_name = self.schema_name(),
                table_name=self.table_name())


    @classmethod
    def from_definition(cls, definition: SchemaMonitorDefinition, workspace):
        # monitor_base = super().from_definition(definition, workspace)
        driver_config = workspace.get_driver_config(definition.datasource)

        columns = [Column.from_dict(col) for col in definition.columns]
        metrics = cls._create_metrics(columns)
        enabled = definition.enabled or True
        
        return cls(
            # name=monitor_base.name,
            # description=monitor_base.description,
            # enabled=monitor_base.enabled,
            # driver_config=monitor_base.driver_config,
            name=definition.name,
            description=definition.description,
            schedule=Schedule(definition.schedule_minutes),
            enabled=enabled,
            driver_config=driver_config,
            table=definition.table,
            columns=columns,
            metrics=metrics,
        )



from flask_restful import Api

from .executions import ExecutionsListResource, ExecutionsResource

def init_api(api):
    # TODO: Provide monitor_id/job_id spaced retrieval
    api.add_resource(ExecutionsListResource, '/v1/api/executions')
    api.add_resource(ExecutionsResource, '/v1/api/execution/<execution_id>')



import logging
from dataclasses import dataclass, field
from typing import Any, List, Dict

from core.monitor.models.schema import SchemaMetric, SchemaMetricType

from .data import Anomaly, Data, Test, TestResult

@dataclass
class SchemaComparisonTest(Test):
    expected: Dict[str, Any]
    data: Dict[str, Any]
    _anomalies: List[Anomaly] = field(default_factory=list)

    @classmethod
    def from_metric(cls, metric: SchemaMetric, schema: Data):
        actual_col_to_metric = {}
        for col in schema.columns:
            actual_col_to_metric[col.name] = getattr(col, metric.type._value_)

        return cls(
            expected=metric.col_to_metric,
            data=actual_col_to_metric,
        )

    def anomalies(self):
        return self._anomalies

    def run(self):
        expected_cols = self.expected.keys()
        actual_cols = self.data.keys()

        for col in expected_cols:
            try:
                if self.expected[col] != self.data[col]:
                    anomaly = Anomaly()
                    self._anomalies.append(anomaly)
            except Exception as e:
                self._anomalies.append(Anomaly())
                logging.warn("Likely a column missing, should be considered an anomaly")

        return TestResult(self._anomalies)




import logging
from dataclasses import dataclass
from typing import Any, Dict, List

from core.common.drivers.column import Table

from core.monitor.models.custom import CustomMetric
from core.monitor.models.schema import SchemaMetric
from core.monitor.models.table import ColumnMetric

from .data import Data
from .schema import SchemaComparisonTest
from .threshold import ThresholdTest
from .zscore import ZScoreTest

class Analyzer:
    def __init__(self, reporter):
        self.reporter = reporter

    def _create_test(self, metric, data: Data):
        if isinstance(metric, ColumnMetric):
            return ZScoreTest.from_metric(metric, data)
        elif isinstance(metric, CustomMetric):
            return ThresholdTest.from_metric(metric, data)
        elif isinstance(metric, SchemaMetric):
            return SchemaComparisonTest.from_metric(metric, data)
        else:
            logging.warning("Data returned from the database did not contain data for a metric defined in the monitor.")
            raise Exception("Could not create test because the monitor was not recognized.")

    def test(self, metric, data):
        test = self._create_test(metric, data)
        
        try:
            self.reporter.test_started(self)
            result = test.run()
            if not len(result.anomalies) > 0:
                self.reporter.test_passed(test)
            else:
                self.reporter.test_failed(test)
        finally:
            self.reporter.test_finished(test)
        
        return result

    def analyze(self, monitor, results):
        data = Data.from_results(results)
        results = {}

        for metric in monitor.retrieve_metrics():
            result = self.test(metric, data)
            results[metric.alias()] = result

        return results



from snowplow_tracker import Tracker, AsyncEmitter, SelfDescribingJson, Subject, logger as snowplow_logger
import logging
import platform

snowplow_logger.setLevel(100)
SNOWPLOW_URL = "monosi-spipeline-collector-lb-1716143593.us-west-2.elb.amazonaws.com" 
# check if dev
e = AsyncEmitter(
    SNOWPLOW_URL, 
    protocol="http",
)
tracker = Tracker(
    e, 
    namespace="cf", 
    app_id="monosi",
)

def set_user_id(user_id):
    subject = Subject()
    subject.set_user_id(user_id)
    tracker.set_subject(subject)

def context():
    # data = {
    #     "name": "platform_context",                
    #     "vendor": platform.platform(),
    #     "version": platform.python_version(),
    #     "format": platform.python_implementation(),
    # }
    return [
        SelfDescribingJson(
            schema="iglu:com.snowplowanalytics.self-desc/instance/jsonschema/1-0-0",
            data={},
            vendor=platform.platform(),
            name="platform_context",
            format=platform.python_implementation(),
            version=platform.python_version(),
        )
    ]

def track_event(config, *args, **kwargs):
    # if not config.send_anonymous_stats:
    #   return
    try:
        tracker.track_struct_event(
            action=kwargs['action'],
            label=kwargs['label'],
            category='monosi',
            # context=context(),
        )
    except Exception as e:
        logging.error("Failed to send anonymous usage stats.")
        logging.error(e)




# TODO: Runner may be responsible for loading
# the configuration for a data source based
# on the monitor definition

class Runner:
    def __init__(self, config):
        self.config = config
        self.driver = None

    def _initialize(self):
        try:
            from core.common.drivers.factory import load_driver
            driver_cls = load_driver(self.config)

            self.driver = driver_cls(self.config)
        except Exception as e:
            print(e)
            raise Exception("Could not initialize connection to database in Runner.")

    def _execute(self, sql: str):
        if self.driver is None:
            raise Exception("Initialize runner before execution.")

        results = self.driver.execute(sql)
        return results

    def run(self, sql: str):
        self._initialize()

        return self._execute(sql)



# from sqlalchemy import Column, DateTime, Integer, func
from sqlalchemy.orm import declarative_base, registry
from server.db import db

# class BaseModel(Base):
#     query_class = MsiQuery

class CrudMixin(object):
    # id = Column(Integer, primary_key=True)
    # updated_at = Column(DateTime(timezone=True), nullable=False, default=func.now())
    # created_at = Column(DateTime(timezone=True), nullable=False, default=func.now())

    # @classmethod
    # def validate(cls, model_dict):
    #     pass

    # @classmethod
    # def from_dict(cls, model_dict):
    #     pass

    # def to_dict(self):
    #     pass

    @classmethod
    def all(cls):
        try:
            return cls.query.all()
        except Exception as e:
            raise Exception("DB: Couldn't query all records")

    def create(self):
        try:
            db.session.add(self)
            db.session.commit()
        except Exception as e:
            raise Exception("DB: Couldn't persist record")

    def update(self, updates):
        try:
            for k, v in updates.items():
                setattr(self, k, v)
            db.session.commit()
        except Exception as e:
            raise Exception("DB: Couldn't update record")

    def delete(self):
        try:
            db.session.delete(self)
            db.session.commit()
        except:
            raise Exception("DB: Couldn't delete record")



from .datasources import (
	DatasourceListResource,
	DatasourceResource,
)
from .integrations import (
    IntegrationListResource,
    IntegrationResource,
)
from .monitors import (
    MonitorListResource,
    MonitorResource,
)

def init_api(api):
	api.add_resource(DatasourceListResource, '/v1/api/datasources')
	api.add_resource(DatasourceResource, '/v1/api/datasources/<obj_id>')

	api.add_resource(IntegrationListResource, '/v1/api/integrations')
	api.add_resource(IntegrationResource, '/v1/api/integrations/<obj_id>')

	api.add_resource(MonitorListResource, '/{}/monitors'.format(api.PREFIX))
	api.add_resource(MonitorResource, '/{}/monitors/<obj_id>'.format(api.PREFIX))



from dataclasses import dataclass
from typing import Any, List

from core.common.drivers.column import Table
from core.common.drivers.base import BaseDialect

from core.monitor.models.metrics import MetricBase
from core.monitor.models.table import TableMonitor, ColumnMetric, ColumnMetricType

@dataclass
class Compiler:
    dialect: BaseDialect
    metadata: Any

    @staticmethod
    def _driver(config):
        try:
            from core.common.drivers.factory import load_driver
            driver_cls = load_driver(config)

            return driver_cls(config)
        except:
            raise Exception("Could not initialize connection to database in runner.")

    def __init__(self, driver_config):
        driver = self._driver(driver_config)
        self.metadata = driver.metadata()
        self.dialect = driver.dialect

    def compile_metric(self, metric: MetricBase):
        return metric.compile(self.dialect)

    def compile_select(self, metrics: List[MetricBase]):
        select_body = []
        for metric in metrics:
            metric_sql = self.compile_metric(metric)
            select_body.append(metric_sql)

        return ",\n\t".join(select_body)

    def _add_cols(self, monitor: TableMonitor):
        tables = Table.from_metadata(self.metadata)
        for table in tables:
            if table.name.lower() in monitor.table.lower():
                monitor.columns = table.columns

    def compile(self, monitor):
        if isinstance(monitor, TableMonitor):
            self._add_cols(monitor)

        select_sql = self.compile_select(monitor.retrieve_metrics())
        sql = monitor.base_sql_statement(select_sql, self.dialect)
        
        return sql




from dataclasses import dataclass
from enum import Enum
from typing import List, Optional


class ColumnDataType(Enum):
    STRING = 'string'
    DATE = 'date'
    DATETIME = 'datetime'
    INTEGER = 'integer'
    FLOAT = 'float'
    BOOLEAN = 'boolean'

def extract_or_default(obj, key, default):
    return obj[key] if key in obj else default

def assign_if_exists(obj, key, value):
    if value:
        obj[key] = value

@dataclass
class Column:
    name: str
    data_type: ColumnDataType
    order: Optional[int] = None
    nullable: Optional[bool] = None
    primary: Optional[bool] = None
    unique: Optional[bool] = None

    @classmethod
    def from_dict(cls, col_dict):
        return cls(
            name=col_dict['name'],
            data_type=ColumnDataType(col_dict['type']),
            order=extract_or_default(col_dict, 'order', None),
            nullable=extract_or_default(col_dict, 'nullable', None),
            primary=extract_or_default(col_dict, 'primary', None),
            unique=extract_or_default(col_dict, 'unique', None),
        )

    def to_dict(self):
        col_dict = {
            'name': self.name,
            'type': self.data_type._value_,
        }
        # TODO: Check if mutable
        assign_if_exists(col_dict, 'order', self.order)
        assign_if_exists(col_dict, 'nullable', self.nullable)
        assign_if_exists(col_dict, 'primary', self.primary)
        assign_if_exists(col_dict, 'unique', self.unique)

        return col_dict

def resolve_to_type_from_str(type_str):
    type_str = type_str.lower()
    if 'varchar' in type_str:
        return ColumnDataType.STRING
    elif 'number' in type_str or 'int' in type_str:
        return ColumnDataType.INTEGER
    elif 'date' in type_str or 'timestamp' in type_str:
        return ColumnDataType.DATE

    return ColumnDataType.STRING

@dataclass
class Table:
    name: str
    columns: List[Column]

    def timestamp_cols(self):
        t_col_types = [ColumnDataType.DATE, ColumnDataType.DATETIME]
        cols = list(filter(lambda x: x.data_type in t_col_types, self.columns))

        return cols

    def timestamp(self):
        return self.timestamp_cols()[0]

    @classmethod
    def _insert_in_table(cls, tables, name, column):
        if name not in tables:
            table = cls(
                name=name,
                columns=[]
            )
        else:
            table = tables[name]

        table.columns.append(column)
        tables[name] = table

        return tables

    @classmethod
    def _fqtn(cls, row, name):
        return '.'.join([row['DATABASE'], row['SCHEMA'], name])

    @classmethod
    def from_metadata(cls, metadata) -> List['Table']:
        tables = {}
        for row in metadata['rows']:
            column = Column(
                name = row['COL_NAME'],
                data_type=resolve_to_type_from_str(row['COL_TYPE']),
                order=int(row['COL_SORT_ORDER'])
            )
            tables = cls._insert_in_table(tables, cls._fqtn(row, row['NAME']), column)

        return list(tables.values()) # TODO: Test me.



from core.common.events import track_event
from core.monitor.models.base import Monitor

from .base import BaseCmd

class RunCmd(BaseCmd):
    def _create_tasks(self):
        return [definition.to_monitor(self.project.configuration) for definition in self.project.monitors]

    def _process_tasks(self):
        track_event(self.project, action="run_start", label="CLI")
        results = [task.run() for task in self.task_queue]
        track_event(self.project, action="run_finish", label="CLI")

        return results


from dataclasses import dataclass
from typing import Any, Dict
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer, String, Table, Text, func
import json

from core.common.drivers.base import BaseDriverConfiguration
from core.monitor.models import MonitorDefinition
from core.monitor.models.schedule import Schedule
from core.monitor.tasks.run import RunMonitorTask

from scheduler.job import MonitorJob
from scheduler.models.execution import Execution

from server.scheduler import manager
from server.integrations import reporter

from . import Base, mapper_registry
from .base import CrudMixin
from .datasource import Datasource


@dataclass
class Workspace:
    datasources: Dict[str, BaseDriverConfiguration]
    name: str = "default"

    @classmethod
    def from_datasource_definitions(cls, datasource_definitions):
        datasources = {}
        for definition in datasource_definitions:
            datasource = cls._config_from_definition(definition)
            datasources[definition.name] = datasource

        return cls(
            datasources=datasources,
        )

    @classmethod
    def _config_from_definition(cls, definition):
        from core.common.drivers.factory import load_config

        try:
            cls = load_config(definition.type)
            ds_config = json.loads(definition.configuration) # TODO: Fix double call
            try:
                ds_config = json.loads(ds_config)
            except:
                pass
            # ds_config = json.loads(definition.configuration)
            # cls.validate(data)
            config = cls.from_dict(ds_config)
        except Exception as e:
            raise e

        return config

    def get_driver_config(self, source_name: str):
        if source_name not in self.datasources.keys():
            raise Exception("Source {} was not defined or can not be found.".format(source_name))

        return self.datasources[source_name]


# TODO: Is this necessary? Is it actually mapped?
# @mapper_registry.mapped

# @dataclass
# class Schedule(Schedule, Base):
#     __table__ = Table(
#         "schedule",
#         mapper_registry.metadata,
#         Column("id", Integer, primary_key=True),
#         Column("minutes", Integer),
#         Column("type", String(50)),
#     )

@dataclass
class Monitor(MonitorDefinition, Base, CrudMixin):
    # TODO: Abstract id, created_at, updated_at
    __table__ = Table(
        "monitor",
        mapper_registry.metadata,
        Column("id", Integer, primary_key=True),

        Column("name", String(50)),
        Column("description", Text),
        Column("type", String(50)),
        Column("enabled", Boolean),
        Column("configuration", Text), # TODO: Better way to not store JSON-style configuration?
        # Column('workspace', String), # TODO: Namespace by workspace
        Column('datasource', String, ForeignKey('datasource.name')),
        # Column("schedule", Text), # TODO: Figure out nested dataclass
        Column("schedule_minutes", Integer),
        Column("schedule_type", String(50)),

        # TODO: Add updated_on fn
        Column("updated_at", DateTime(timezone=True), nullable=False, server_default=func.now()),
        Column("created_at", DateTime(timezone=True), nullable=False, server_default=func.now()),
    )
    # TODO: https://stackoverflow.com/questions/61108811/sqlalchemy-relationship-with-foreignkey-consisting-of-two-columns

    # source = relationship() # TOOD: Declare
    # workspace = relationship()

    def to_dict(self):
        obj_dict = super().to_dict()

        obj_dict['id'] = self.id
        try:
            obj_dict['last_run'] = str(Execution.get_by_job_id(self.id).created_at)
        except:
            pass
        obj_dict['updated_at'] = str(self.updated_at)
        obj_dict['created_at'] = str(self.created_at)

        return obj_dict
        
    def run(self):
        workspace = Workspace.from_datasource_definitions(Datasource.all())
        monitor = self.to_monitor(workspace)
        results = monitor.run()
        print("{} successfully ran.".format(self.name))

    def schedule(self):
        try:
            workspace = Workspace.from_datasource_definitions(Datasource.all())
            monitor = self.to_monitor(workspace)
            task = RunMonitorTask(monitor)
            task.reporter = reporter
            job = MonitorJob(task)
            manager.add_job(job, job_id=self.id)
        except Exception as e:
            raise e

    def create(self):
        result = super().create()
        self.schedule()
        # self.run()

    def delete(self):
        manager.remove_job(str(self.id))
        Execution.delete_by_job_id(self.id)
        return super().delete()

    # TODO: Implementation - should remove route while pending
    # def update(self):
    #     pass



# TODO: Polymorphic association
    # __mapper_args__ = {
    #     'polymorphic_identity':'monitor',
    #     'polymorphic_on':monitor_type,
    #     'with_polymorphic': '*'
    # }



from dataclasses import dataclass
import json
from sqlalchemy import Boolean, Column, DateTime, Integer, String, Table, Text, func

from core.common.drivers.base import DriverDefinition

from . import Base, mapper_registry
from .base import CrudMixin

@dataclass
class TempDatasourceDefinition:
    name: str
    type: str
    configuration: str

    def to_dict(self):
        return {
            "name": self.name,
            "type": self.type,
            "configuration": self.configuration,
        }

    @classmethod
    def from_dict(cls, ds_dict):
        return cls(
            name=ds_dict['name'],
            type=ds_dict['type'],
            configuration=json.dumps(ds_dict['configuration'])
        )

# TODO: Is this necessary? Is it actually mapped?
# @mapper_registry.mapped

@dataclass
class Datasource(DriverDefinition, Base, CrudMixin):
    # TODO: Abstract id, created_at, updated_at
    __table__ = Table(
        "datasource",
        mapper_registry.metadata,
        Column("id", Integer, primary_key=True),
        Column("name", String(50), unique=True),
        Column("type", String(50)), # TODO: Enum
        Column("configuration", Text), # TODO: Better way to not store JSON-style configuration?

        # TODO: Add updated_on fn
        Column("updated_at", DateTime(timezone=True), nullable=False, server_default=func.now()),
        Column("created_at", DateTime(timezone=True), nullable=False, server_default=func.now()),
    )

    def to_dict(self):
        obj_dict = super().to_dict()

        obj_dict['id'] = self.id
        obj_dict['updated_at'] = str(self.updated_at)
        obj_dict['created_at'] = str(self.created_at)

        return obj_dict

    def create(self):
        configuration = json.dumps(self.configuration)
        self.configuration = configuration
        super().create()

    # def profile(self):
    #     pass



from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Optional

from core.common.drivers.column import Column, ColumnDataType
from core.common.drivers.base import BaseDialect

from . import MonitorDefinition, MonitorConfiguration
from .base import Monitor
from .metrics import MetricBase
from .schedule import Schedule

@dataclass
class TableMonitorConfigurationDefaults:
    where: str = ''
    days_ago: int = -100

@dataclass
class TableMonitorConfiguration:
    table: str
    timestamp_field: str

@dataclass
class TableMonitorDefinition(MonitorConfiguration, TableMonitorConfigurationDefaults, MonitorDefinition, TableMonitorConfiguration):
    def to_monitor(self, workspace):
        return TableMonitor.from_definition(self, workspace)

    def to_dict(self):
        return {
            "name": self.name,
            "description": self.description,
            "type": "table",
            "table": self.table,
            # "schedule_minutes": self.schedule_minutes,
            # "schedule_type": self.schedule_type,
            # "schedule": Schedule(self.schedule_minutes).to_dict(),
            "timestamp_field": self.timestamp_field,
        }

def extract_or_default(obj, key, default):
    return obj[key] if key in obj else default

class ColumnMetricType(Enum):
    APPROX_DISTINCTNESS = 'approx_distinctness'
    COMPLETENESS = 'completeness'
    ZERO_RATE = 'zero_rate'
    NEGATIVE_RATE = 'negative_rate'
    NUMERIC_MEAN = 'numeric_mean'
    NUMERIC_MIN = 'numeric_min'
    NUMERIC_MAX = 'numeric_max'
    NUMERIC_STD = 'numeric_std'
    APPROX_DISTINCT_COUNT = 'approx_distinct_count'
    MEAN_LENGTH = 'mean_length'
    MAX_LENGTH = 'max_length'
    MIN_LENGTH = 'min_length'
    STD_LENGTH = 'std_length'
    TEXT_INT_RATE = 'text_int_rate'
    TEXT_NUMBER_RATE = 'text_number_rate'
    TEXT_UUID_RATE = 'text_uuid_rate'
    TEXT_ALL_SPACES_RATE = 'text_all_spaces_rate'
    TEXT_NULL_KEYWORD_RATE = 'text_null_keyword_rate'

    @classmethod
    def default(cls):
        return [cls.APPROX_DISTINCTNESS, cls.COMPLETENESS]

    @classmethod
    def all(cls):
        return [
            cls.APPROX_DISTINCTNESS, 
            cls.COMPLETENESS,
            cls.ZERO_RATE,
            cls.NEGATIVE_RATE,
            cls.NUMERIC_MEAN,
            cls.NUMERIC_MIN,
            cls.NUMERIC_MAX,
            cls.NUMERIC_STD,
            cls.APPROX_DISTINCT_COUNT,
            cls.MEAN_LENGTH,
            cls.MAX_LENGTH,
            cls.MIN_LENGTH,
            cls.STD_LENGTH,
            cls.TEXT_INT_RATE,
            cls.TEXT_NUMBER_RATE,
            cls.TEXT_UUID_RATE,
            cls.TEXT_ALL_SPACES_RATE,
            cls.TEXT_NULL_KEYWORD_RATE,
        ]

    @classmethod
    def default_for(cls, data_type: ColumnDataType): 
        if data_type == ColumnDataType.FLOAT or data_type == ColumnDataType.INTEGER:
            return [
                cls.ZERO_RATE,
                cls.NEGATIVE_RATE,
                cls.NUMERIC_MEAN,
                cls.NUMERIC_MIN,
                cls.NUMERIC_MAX,
                cls.NUMERIC_STD,
            ]
        elif data_type == ColumnDataType.STRING:
            return [
                cls.APPROX_DISTINCT_COUNT,
                cls.MEAN_LENGTH,
                cls.MAX_LENGTH,
                cls.MIN_LENGTH,
                cls.STD_LENGTH,
                cls.TEXT_INT_RATE,
                cls.TEXT_NUMBER_RATE,
                cls.TEXT_UUID_RATE,
                cls.TEXT_ALL_SPACES_RATE,
                cls.TEXT_NULL_KEYWORD_RATE,
            ]
            

        return cls.default()


@dataclass
class ColumnMetric(MetricBase):
    type: ColumnMetricType
    column: str

    def alias(self):
        return "{}__{}".format(self.column, self.type.value)

    def compile(self, dialect: BaseDialect):
        alias = self.alias()
        column = self.column

        attr = getattr(dialect, self.type._value_)
        if not attr:
            raise Exception("Unreachable: Metric type is defined that does not resolve to a definition.")
        unformatted_metric_sql = attr()
        formatted_metric_sql = unformatted_metric_sql.format(column)

        metric_sql = "{} AS {}".format(formatted_metric_sql, alias)
        return metric_sql


@dataclass
class TableMonitor(TableMonitorConfigurationDefaults, Monitor, TableMonitorConfiguration):
    metrics: List[ColumnMetricType] = field(default_factory=lambda: ColumnMetricType.all())
    columns: List[Column] = field(default_factory=list)
    minutes_ago: int = -100 * 24 * 60

    def base_sql_statement(self, select_sql, dialect):
        return dialect.table_query().format(
            select_sql=select_sql,
            table=self.table,
            timestamp_field=self.timestamp_field,
            minutes_ago=self.minutes_ago,
        )

    def info(self):
        info_str = "Table Health Monitor: {}".format(self.table)
        if self.description:
            info_str += "\n{}".format(self.description)
        return info_str

    @classmethod
    def validate(cls, monitor_dict):
        pass

    def _create_metrics(self):
        metrics = []
        for column in self.columns:
            for metric_type in ColumnMetricType.default_for(column.data_type):
                if metric_type in self.metrics:
                    metric = ColumnMetric(
                        column=column.name,
                        type=ColumnMetricType(metric_type),
                    )
                    metrics.append(metric)

        return metrics
    
    @classmethod
    def from_dict(cls, value: Dict[str, Any]) -> 'Monitor':
        table = value['table']
        timestamp_field = value['timestamp_field']
        description = extract_or_default(value, 'description', None)
        where = extract_or_default(value, 'where', '')
        days_ago = extract_or_default(value, 'days_ago', -100)
        
        return cls(
            table=table,
            description=description,
            timestamp_field=timestamp_field,
            where=where,
            minutes_ago=(days_ago * 60 * 24),
        )

    def to_dict(self):
        # Used for output on bootstrap right now.
        return {
            'table': self.table,
            'timestamp_field': self.timestamp_field,
            'type': 'table',
        }

    def retrieve_metrics(self):
        return self._create_metrics()

    @classmethod
    def from_definition(cls, definition: TableMonitorDefinition, workspace):
        # monitor_base = super().from_definition(definition, workspace)
        driver_config = workspace.get_driver_config(definition.datasource)

        return cls(
            # name=monitor_base.name,
            # description=monitor_base.description,
            # enabled=monitor_base.enabled,
            # driver_config=monitor_base.driver_config,
            name=definition.name,
            description=definition.description,
            schedule=Schedule(definition.schedule_minutes),
            enabled=definition.enabled or True,
            driver_config=driver_config,
            table=definition.table,
            timestamp_field=definition.timestamp_field,
            where=definition.where,
            minutes_ago=(definition.days_ago * 60 * 24),
        )




from dataclasses import dataclass

from core.common.drivers.base import BaseDriverConfiguration

@dataclass
class SnowflakeDriverConfiguration(BaseDriverConfiguration):
    user: str
    password: str
    warehouse: str
    database: str
    # schema: str
    account: str

    def driver_name(self):
        return "snowflake"

    def configuration_schema(self):
        return {}

    def connection_string(self) -> str:
        # /<database_name>/<schema_name>?warehouse=<warehouse_name>&role=<role_name>'
        return 'snowflake://{user}:{password}@{account}/{database}?warehouse={warehouse}'.format(
            user=self.user,
            password=self.password,
            account=self.account,
            database=self.database,
            warehouse=self.warehouse,
        )

    @classmethod
    def from_dict(cls, config_dict):
        return cls(
            user=config_dict.get('user'),
            password=config_dict.get('password'),
            warehouse=config_dict.get('warehouse'),
            database=config_dict.get('database'),
            account=config_dict.get('account'),
        )

    def to_dict(self):
        return {
            'user': self.user,
            'warehouse': self.warehouse,
            'database': self.database,
            'account': self.account,
        }
    



from server.models.datasource import Datasource
from .base import CrudResource, ListResource

class DatasourceListResource(ListResource):
    @property
    def resource(self):
        return Datasource

    @property
    def key(self):
        return "datasources"

class DatasourceResource(CrudResource):
    @property
    def resource(self):
        return Datasource

    @property
    def key(self):
        return "datasource"




from server.models.monitor import Monitor
from .base import CrudResource, ListResource

class MonitorListResource(ListResource):
    @property
    def resource(self):
        return Monitor

    @property
    def key(self):
        return "monitors"

class MonitorResource(CrudResource):
    @property
    def resource(self):
        return Monitor

    @property
    def key(self):
        return "monitor"



from core.common.drivers.base import BaseDialect

class PostgresDialect(BaseDialect):
    @classmethod
    def numeric_std(cls):
        return "STDDEV(CAST({} as double precision))"
    
    @classmethod
    def text_int_rate(cls):
        return "SUM(CASE WHEN CAST({} AS varchar) ~ '^([-+]?[0-9]+)$' THEN 1 ELSE 0 END) / CAST(COUNT(*) AS NUMERIC)"

    @classmethod
    def text_number_rate(cls):
        return "SUM(CASE WHEN CAST({} AS varchar) ~ '^([-+]?[0-9]*[.]?[0-9]+([eE][-+]?[0-9]+)?)$' THEN 1 ELSE 0 END) / CAST(COUNT(*) AS NUMERIC)"

    @classmethod
    def text_uuid_rate(cls):
        return "SUM(CASE WHEN CAST({} AS varchar) ~ '^([0-9a-fA-F]{{8}}-[0-9a-fA-F]{{4}}-[0-9a-fA-F]{{4}}-[0-9a-fA-F]{{4}}-[0-9a-fA-F]{{12}})$' THEN 1 ELSE 0 END) / CAST(COUNT(*) AS NUMERIC)"

    @classmethod
    def text_all_spaces_rate(cls):
        return "SUM(CASE WHEN CAST({} AS varchar) ~ '^(\\\\s+)$' THEN 1 ELSE 0 END) / CAST(COUNT(*) AS NUMERIC)"

    @classmethod
    def text_null_keyword_rate(cls):
        return "SUM(CASE WHEN UPPER(CAST({} as varchar)) IN ('NULL', 'NONE', 'NIL', 'NOTHING') THEN 1 ELSE 0 END) / CAST(COUNT(*) AS NUMERIC)"

    @classmethod
    def zero_rate(cls): # TODO: ?
        return "SUM(CASE WHEN UPPER(CAST({} as varchar)) IN ('NULL', 'NONE', 'NIL', 'NOTHING') THEN 1 ELSE 0 END) / CAST(COUNT(*) AS NUMERIC)"

    @classmethod
    def negative_rate(cls):
        return "SUM(CASE WHEN {} < 0 THEN 1 ELSE 0 END) / CAST(COUNT(*) AS NUMERIC)"

    @classmethod
    def completeness(cls):
        return "COUNT({}) / CAST(COUNT(*) AS NUMERIC)"

    @classmethod
    def metadata_query(cls):
        return """
        SELECT
            lower(c.table_name) AS name,
            lower(c.column_name) AS COL_NAME,
            lower(c.data_type) AS COL_TYPE,
            ordinal_position AS COL_SORT_ORDER, 
            lower(c.table_catalog) AS database,
            lower(c.table_schema) AS schema
        FROM
            INFORMATION_SCHEMA.COLUMNS AS c
        LEFT JOIN
            INFORMATION_SCHEMA.TABLES t
                ON c.TABLE_NAME = t.TABLE_NAME
                AND c.TABLE_SCHEMA = t.TABLE_SCHEMA
        WHERE LOWER( c.table_name ) = '{table_name}'
          AND LOWER( c.table_schema ) = '{schema_name}'
          AND LOWER( c.table_catalog ) = '{database_name}'"""

    @classmethod
    def table_query(cls):
        return """
            SELECT 
                DATE_TRUNC('HOUR', {timestamp_field}) as window_start, 
                DATE_TRUNC('HOUR', {timestamp_field}) + interval '1 hour' as window_end,

                COUNT(*) as row_count,

                {select_sql}

            FROM {table}
            WHERE 
                DATE_TRUNC('HOUR', {timestamp_field}) >= CURRENT_TIMESTAMP + interval '{minutes_ago} minutes' 
            GROUP BY window_start, window_end 
            ORDER BY window_start ASC;
        """



from dataclasses import dataclass, field
from enum import Enum
from typing import Optional 
import json

from .schedule import Schedule, ScheduleType

def load_monitor_cls(monitor_dict):
    monitor_type = monitor_dict.get('type').lower()
    # monitor_type = MonitorType(type_raw)

    if monitor_type == "table":
        from .table import TableMonitor
        return TableMonitor
    elif monitor_type == "custom":
        from .custom import CustomMonitor
        return CustomMonitor
    elif monitor_type == "schema":
        from .schema import SchemaMonitor
        return SchemaMonitor

    raise Exception("Could not find a moniotr with type: {}".format(monitor_type))

def load_monitor_definition(monitor_definition):
    monitor_type = monitor_definition['type']
    # monitor_type = MonitorType(type_raw)

    configuration = monitor_definition['configuration']
    if monitor_type == "table":
        from .table import TableMonitorDefinition
        # TableMonitorDefinition.validate()
        return TableMonitorDefinition(
            **monitor_definition,
            **configuration,
        )
    elif monitor_type == "custom":
        from .custom import CustomMonitorDefinition
        # CustomMonitorDefinition.validate()
        return CustomMonitorDefinition(
            **monitor_definition,
            **configuration,
        )
    elif monitor_type == "schema":
        from .schema import SchemaMonitorDefinition
        # SchemaMonitorDefinition.validate()
        return SchemaMonitorDefinition(
            **monitor_definition,
            **configuration,
        )

    # Note: Unreachable - we would error at MonitorType instantiation
    raise Exception("Could not find a monitor definition with type: {}".format(monitor_type))

@dataclass
class MonitorConfiguration:
    pass

# TODO: It's likely that a monitor definition
# will have to include information about which
# datasource it is intended to run on and a
# way to grab its configuration information
@dataclass
class MonitorDefinition:
    name: str
    type: str
    # workspace: str
    configuration: str # TODO (This is NOT a connection configuration, it is the details for the individ monitor)
    schedule_minutes: int = 720
    schedule_type: str = ScheduleType.INTERVAL._value_
    datasource: str = 'default' # Used (with workspace) to resolve the DriverConfig
    description: Optional[str] = None
    enabled: Optional[bool] = True
    # schedule: Schedule = field(default_factory=lambda: Schedule())

    @classmethod
    def validate(cls, monitor_dict):
        pass

    def to_dict(self):
        return {
            "name": self.name,
            "description": self.description,
            "enabled": self.enabled,
            "type": self.type,
            "datasource": self.datasource,
            # "schedule": Schedule(self.schedule_minutes).to_dict(),
            "schedule_minutes": self.schedule_minutes,
            "schedule_type": self.schedule_type,
            "configuration": json.loads(self.configuration),
        }

    @classmethod
    def from_dict(cls, monitor_dict):
        if 'configuration' in monitor_dict:
            configuration = json.dumps(monitor_dict['configuration'])
            monitor_dict['configuration'] = configuration
        else:
            monitor_dict['configuration'] = '{}'

        return cls(
            **monitor_dict,
        )

    def to_monitor(self, workspace): # TODO: to_dict contains id, updated_at, created_at
        monitor_dict = self.to_dict()
        monitor_dict.pop('id')
        try:
            monitor_dict.pop('last_run')
        except:
            pass
        monitor_dict.pop('created_at')
        monitor_dict.pop('updated_at')
        monitor_cls = load_monitor_definition(monitor_dict)
        return monitor_cls.to_monitor(workspace)




from flask import Flask, send_from_directory
from flask_cors import CORS
import os
import sys
import uuid

from core.common.events import set_user_id

from .api import MsiApi
from .db import db
from .models.monitor import Monitor
from .models.integration import Integration
from .scheduler import manager

curr_path = os.path.dirname(os.path.abspath(__file__))
src_path = os.path.abspath(os.path.join(curr_path, "../"))
sys.path.append(src_path)

def _build_path():
    build_path = os.path.join(src_path, 'ui/build/')
    if not os.path.exists(build_path):
        raise Exception("Client UI was not built before attempting to serve via Flask.")

    return build_path

def _serve_ui(path=''):
    build_path = _build_path()
    req_path = os.path.join(build_path, path)

    if req_path == build_path or not os.path.exists(req_path):
        path = "index.html"

    return send_from_directory(build_path, path)

def _init_ui(app):
    if not app.config['SERVE_UI']:
        return

    app.static_folder =  os.path.join(_build_path(), "static")
    app.add_url_rule("/", view_func=_serve_ui)
    app.add_url_rule("/<path:path>", view_func=_serve_ui)


def create_app():
    # Initialize Server
    app = Flask(__name__)
    CORS(app, resources={r"/v1/api/*": {"origins": "*"}})
    app_settings = os.getenv(
        'APP_SETTINGS',
        'server.config.Config'
    )
    app.config.from_object(app_settings)

    # Initialize UI if SERVE_UI
    _init_ui(app)

    # Initialize DB
    db.init_app(app)
    db.app = app
    with app.app_context():
        db.create_all()

    # Initialize API
    api = MsiApi(app)

    # Initialize monitor jobs
    manager.init_app(app)
    with app.app_context():
        [monitor.schedule() for monitor in Monitor.all()]

    [integration.register() for integration in Integration.all()]

    set_user_id(str(uuid.uuid4()))

    return app



from dataclasses import dataclass

from core.common.drivers.base import BaseDriverConfiguration

@dataclass
class PostgresDriverConfiguration(BaseDriverConfiguration):
    user: str
    password: str
    host: str
    database: str
    port: int = 5432

    def driver_name(self):
        return "postgres"

    def connection_string(self) -> str:
        return 'postgresql://{user}:{password}@{host}:{port}/{database}'.format(
            user=self.user,
            password=self.password,
            host=self.host,
            port=self.port,
            database=self.database,
        )

    @classmethod
    def from_dict(cls, config_dict):
        return cls(
            user=config_dict.get('user'),
            password=config_dict.get('password'),
            database=config_dict.get('database'),
            host=config_dict.get('host'),
            port=config_dict.get('port'),
        )

    def to_dict(self):
        return {
            'user': self.user,
            'database': self.database,
            'host': self.host,
            'port': self.port,
        }
    


from datetime import datetime, timedelta
from flask_apscheduler import APScheduler
import json
import uuid
import logging

# from monosi.scheduler.db import db
import scheduler.constants as constants
from .models.execution import Execution

logging.getLogger('apscheduler').setLevel(logging.DEBUG)

def import_from_path(path):
    components = path.split('.')
    module = __import__('.'.join(components[:-1]))
    for comp in components[1:-1]:
        module = getattr(module, comp)
    return getattr(module, components[-1])

class MsiScheduler(APScheduler):
    @classmethod
    def run_job(cls, func, job_id, *args, **kwargs):
        execution = Execution(
            job_id=job_id,
            state=constants.STATUS_SCHEDULED
        )
        execution.create()
        execution_id = execution.id

        try:
            execution.update(updates={"state": constants.STATUS_SCHEDULED})

            cls.run_scheduler_job(func, execution_id, *args, **kwargs)
        except Exception:
            execution.update(updates={"state": constants.STATUS_SCHEDULED_ERROR})

            return None
        return execution_id

    @classmethod
    def run_scheduler_job(cls, func, execution_id, *args, **kwargs):
        try:
            execution = Execution.get_by_id(execution_id)
            execution.update(updates={"state": constants.STATUS_RUNNING})

            result = func()
            result_json = json.dumps(result, indent=4, sort_keys=True)

            execution.update(updates={
                "state": constants.STATUS_SUCCEEDED,
                "result": result_json
            })

        except Exception as err:
            print("Error: {0}".format(err))
            if execution:
                execution.update({
                    "state": constants.STATUS_FAILED
                })


    def add_scheduler_job(self, monitor_job, job_id=None, args=None, trigger='interval', minutes=720, **kwargs):
        if not job_id:
            job_id = str(uuid.uuid4().hex)

        execution = Execution.get_by_job_id(job_id)
        if execution:
            monitor_job.task.monitor.minutes_ago = int((datetime.now() - execution.created_at).seconds / 60)

        if not args:
            args = []
        arguments = [monitor_job.run, job_id]
        arguments.extend(args)

        start_date = datetime.now() - timedelta(minutes=(minutes-1)) # start one minute after schedule

        self.add_job(func=self.run_job, id=job_id, args=arguments, trigger=trigger, minutes=minutes, start_date=start_date)

        return job_id



from core.common.drivers.base import BaseDialect

class SnowflakeDialect(BaseDialect):
    @classmethod
    def text_int_rate(cls):
        return "SUM(IFF(REGEXP_COUNT(TO_VARCHAR({}), '^([-+]?[0-9]+)$', 1, 'i') != 0, 1, 0)) / CAST(COUNT(*) AS NUMERIC)"

    @classmethod
    def text_number_rate(cls):
        return "SUM(IFF(REGEXP_COUNT(TO_VARCHAR({}), '^([-+]?[0-9]*[.]?[0-9]+([eE][-+]?[0-9]+)?)$', 1, 'i') != 0, 1, 0)) / CAST(COUNT(*) AS NUMERIC)"

    @classmethod
    def text_uuid_rate(cls):
        return "SUM(IFF(REGEXP_COUNT(TO_VARCHAR({}), '^([0-9a-fA-F]{{8}}-[0-9a-fA-F]{{4}}-[0-9a-fA-F]{{4}}-[0-9a-fA-F]{{4}}-[0-9a-fA-F]{{12}})$', 1, 'i') != 0, 1, 0)) / CAST(COUNT(*) AS NUMERIC)"

    @classmethod
    def text_all_spaces_rate(cls):
        return "SUM(IFF(REGEXP_COUNT(TO_VARCHAR({}), '^(\\\\s+)$', 1, 'i') != 0, 1, 0)) / CAST(COUNT(*) AS NUMERIC)"

    @classmethod
    def text_null_keyword_rate(cls):
        return "SUM(IFF(UPPER({}) IN ('NULL', 'NONE', 'NIL', 'NOTHING'), 1, 0)) / CAST(COUNT(*) AS NUMERIC)"

    @classmethod
    def zero_rate(cls):
        return "SUM(IFF(UPPER({}) IN ('NULL', 'NONE', 'NIL', 'NOTHING'), 1, 0)) / CAST(COUNT(*) AS NUMERIC)"

    @classmethod
    def negative_rate(cls):
        return "SUM(IFF({} < 0, 1, 0)) / CAST(COUNT(*) AS NUMERIC)"

    @classmethod
    def completeness(cls):
        return "COUNT({}) / CAST(COUNT(*) AS NUMERIC)"

    @classmethod
    def metadata_query(cls):
        return """
            SELECT
                lower(c.table_name) AS name,
                lower(c.column_name) AS col_name,
                lower(c.data_type) AS col_type,
                c.comment AS col_description,
                lower(c.ordinal_position) AS col_sort_order,
                lower(c.table_catalog) AS database,
                lower(c.table_schema) AS schema,
                t.comment AS description,
                decode(lower(t.table_type), 'view', 'true', 'false') AS is_view
            FROM
                {database_name}.INFORMATION_SCHEMA.COLUMNS AS c
            LEFT JOIN
                {database_name}.INFORMATION_SCHEMA.TABLES t
                    ON c.TABLE_NAME = t.TABLE_NAME
                    AND c.TABLE_SCHEMA = t.TABLE_SCHEMA
            WHERE LOWER( name ) = '{table_name}'
              AND LOWER( schema ) = '{schema_name}'
        """

    @classmethod
    def table_query(cls):
        return """
            SELECT 
                DATE_TRUNC('HOUR', {timestamp_field}) as window_start, 
                DATEADD(hr, 1, DATE_TRUNC('HOUR', {timestamp_field})) as window_end, 

                COUNT(*) as row_count, 

                {select_sql}

            FROM {table}
            WHERE 
                DATE_TRUNC('HOUR', {timestamp_field}) >= DATEADD(minute, {minutes_ago}, CURRENT_TIMESTAMP()) 
            GROUP BY window_start, window_end 
            ORDER BY window_start ASC;
        """




from flask_restful import Api

import server.handlers as handlers

class MsiApi(Api):
    VERSION = 'v1'
    PREFIX = f'{VERSION}/api'

    def __init__(self, app):
        super().__init__(app)
        handlers.init_api(self)



from core.common.reporter import Reporter

class RunMonitorTask:
    def __init__(self, monitor):
        self.monitor = monitor
        self.reporter = Reporter()

    # def _retrieve_driver_config(self):
    #     try:
    #         from core.common.drivers.factory import load_driver_config
    #         driver_config = self.monitor.driver_config
    #         driver_config_cls = load_driver_config(driver_config)

    #         return driver_config_cls(self.config)
    #     except:
    #         raise Exception("Could not initialize connection to database in runner.")

    def _retrieve_driver_config(self):
        return self.monitor.driver_config

    def _compile(self):
        from core.monitor.compiler import Compiler

        compiler = Compiler(self._retrieve_driver_config())
        return compiler.compile(self.monitor)

    def _run(self, sql):
        from core.monitor.runner import Runner

        runner = Runner(self._retrieve_driver_config())
        return runner.run(sql)

    def _analyze(self, results):
        from core.monitor.analyzer import Analyzer
        
        analyzer = Analyzer(self.reporter)
        return analyzer.analyze(self.monitor, results)

    def run(self):
        self.reporter.monitor_started(self.monitor)

        compiled_sql = self._compile()
        runner_results = self._run(compiled_sql)
        analysis = self._analyze(runner_results)

        self.reporter.monitor_finished(self.monitor)
        # analysis = self._analyze(runner_results)
        return runner_results

        # return (runner_results, analysis)

# class RunAllMonitorsTask:
#     def __init__(self, monitors):
#         self.monitors = monitors

#     def _create_tasks(self):
#         return self.monitors

#     @staticmethod
#     def _process_tasks(tasks):
#         tasks = self._create_tasks()
#         return [task.run() for task in tasks]
    
#     def run(self):
#         return self._process_tasks()
